#!/usr/bin/env python
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2012 Brian Hahn, 2020 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>Path Copy Rotate Shift Plus


from gimpfu import *
#from array import array
#import math
import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
#darkgrey = gimpcolor.RGB(50, 50, 50)
#midgrey = gimpcolor.RGB(125, 125, 125)
#lightgrey = gimpcolor.RGB(170, 170, 170)
#verylightgrey = gimpcolor.RGB(222, 222, 222)
#red = gimpcolor.RGB(220,10,10)	



def do_stuff(image, layer, pf_offset_x, pf_offset_y, pf_angle, pf_centre_X, pf_centre_Y, use_centre, num_copies, path_width, modifier, modifier_options, colour_option, path_options):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    
    layer_list = image.layers
    num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
	
    if num_paths>0:
	
        angle_increment, increment, step = 0, 0, 0
        increment = modifier
		
        active_vector = pdb.gimp_image_get_active_vectors(image)
        if active_vector == -1:
            pdb.gimp_image_set_active_vectors(image, vector_list[0])
            active_vector = pdb.gimp_image_get_active_vectors(image)
        copy_vector = active_vector
			
        pdb.gimp_item_set_visible(active_vector, False)
			
        active_layer = pdb.gimp_image_get_active_layer(image)
        if active_layer == -1: active_layer = pdb.gimp_image_set_active_layer(image, layer_list[0])
		
		
        pdb.gimp_context_set_stroke_method(0) # Use line with stroke method
		
        # get current colours
        original_fg_colour = pdb.gimp_context_get_foreground()
        original_bg_colour = pdb.gimp_context_get_background()
        original_current_gradient = pdb.gimp_context_get_gradient()
        current_gradient = original_current_gradient
        # get gradient colour information
        list_of_gradient_positions = fn_get_gradient_positions(num_copies)
        number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
                
			
        #pdb.gimp_message(str(active_vector) )
            
        if use_centre: # calculate image centre for initial position
            pf_centre_X = pdb.gimp_image_width(image)/2
            pf_centre_Y = pdb.gimp_image_height(image)/2
			
        loop = 1
        while loop<=num_copies:
            new_vector = pdb.gimp_vectors_copy(copy_vector) # copy the user supplied vector
            pdb.gimp_image_insert_vectors(image, new_vector, None, -1)
            new_vector_name = "skinnypath " + str((loop))
            pdb.gimp_item_set_name(new_vector, new_vector_name)
			
            if modifier_options == 0:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, loop * pf_angle)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x) ), loop * pf_offset_y)
            elif modifier_options == 1:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, loop * pf_angle)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x) + increment ), loop * pf_offset_y)
            elif modifier_options == 2:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, loop * pf_angle)
                pdb.gimp_vectors_stroke_translate(new_vector, True, loop * pf_offset_x , (loop * pf_offset_y) + increment )
            elif modifier_options == 3:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, (loop * pf_angle) + increment)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x) ), loop * pf_offset_y)
            elif modifier_options == 4:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, loop * pf_angle)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x)  + increment), (loop * pf_offset_y) + increment)
            elif modifier_options == 5:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, (loop * pf_angle) + increment)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x) + increment ), loop * pf_offset_y)
            elif modifier_options == 6:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, (loop * pf_angle) + increment)
                pdb.gimp_vectors_stroke_translate(new_vector, True, loop * pf_offset_x , (loop * pf_offset_y) + increment )
            elif modifier_options == 7:
                pdb.gimp_vectors_stroke_rotate(new_vector, True, pf_centre_X, pf_centre_Y, (loop * pf_angle) + increment)
                pdb.gimp_vectors_stroke_translate(new_vector, True, ((loop * pf_offset_x)  + increment), (loop * pf_offset_y) + increment)
                
            
			
            pdb.gimp_context_set_line_width(float(path_width))
			
            # colour options
            if colour_option == 0: # foreground
                path_colour = original_fg_colour # all paths set to fg colour
            elif colour_option == 1: # background
                path_colour = original_bg_colour # all paths set to bg colour
            elif colour_option == 2: # foreground to background blend
                path_colour = fn_shade_colour(original_fg_colour, original_bg_colour, num_copies, loop)
            elif colour_option == 3: # background to foreground blend
                path_colour = fn_shade_colour(original_bg_colour, original_fg_colour, num_copies, loop)
            elif colour_option == 4: # use gradient colours in sequence 
                path_colour = fn_shade_gradient_colour(colours_of_samples,loop)
            elif colour_option == 5: # use random gradient colours
                random.seed()
                col_loop = random.randrange(0, num_copies)
                path_colour = fn_shade_gradient_colour(colours_of_samples,col_loop)	
                

            pdb.gimp_context_set_foreground(path_colour)
			
            # stroke the path
            pdb.gimp_drawable_edit_stroke_item(active_layer, new_vector) # stroke the path
            step += modifier
            if modifier <> 0.0: increment += step
            #pdb.gimp_message("distance_increment = " + str(distance_increment) + " (loop * pf_offset_x) = " + str((loop * pf_offset_x) ) )
            #pdb.gimp_message("(loop * pf_offset_x) + distance_increment = " + str((loop * pf_offset_x) + distance_increment) )
			
            # options for dealing with the generated paths
            if path_options == 1: # Merge all paths into one - apart from original user path
                if loop>1: 
            
                    num_merge_strokes, merge_vector_list = pdb.gimp_vectors_get_strokes(new_vector)
                    merge_vector_id = new_vector
                    stroke_loop = 0
                    while stroke_loop<num_merge_strokes:
                        # loop through paths that make up the new vector
                        merge_vector_stroke = merge_vector_list[stroke_loop]
                        # get information for the most recent path
                        TypeOfStrokeInPart_Vector,NumberOfPointsInMerge_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(merge_vector_id, merge_vector_stroke)
                        merge_ListOfControlPoints = ListOfControlPoints
                        # add new path strokes to the merge path
                        new_vector = pdb.gimp_vectors_stroke_new_from_points(old_vector, 0, NumberOfPointsInMerge_Vector, ListOfControlPoints, IsStrokeClosed)
							
                        stroke_loop += 1
                    		
                    		
                if loop == 1: old_vector = new_vector # this vector will become the merged path 
                if loop>=2: 
                    active_vector = pdb.gimp_image_get_active_vectors(image)
                    pdb.gimp_image_remove_vectors(image, active_vector)	# remove latest path after merging
            if path_options == 2: pdb.gimp_image_remove_vectors(image, new_vector)
			
			
            loop += 1
	
	
    else:
        message_text = "Please add a path to this image and try again."
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)  
	
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

    #### Useful Functions
	
######################################################################################################################################

def fn_shade_gradient_colour(colours_of_samples, instances):
    red_amount = colours_of_samples[(instances*4) + 0] * 255
    green_amount = colours_of_samples[(instances*4) + 1] * 255
    blue_amount = colours_of_samples[(instances*4) + 2] * 255
    alpha_amount = colours_of_samples[(instances*4) + 3] * 255

    if red_amount < 0: red_amount = 0
    if red_amount > 255: red_amount = 255
    if green_amount < 0: green_amount = 0
    if green_amount > 255: green_amount = 255
    if blue_amount < 0: blue_amount = 0
    if blue_amount > 255: blue_amount = 255
    if alpha_amount < 0: alpha_amount = 0
    if alpha_amount > 255: alpha_amount = 255
	
    new_f_colour = (int(red_amount), int(green_amount), int(blue_amount), int(alpha_amount))
    return new_f_colour
	
#######################################################################################################

def fn_get_gradient_positions(instances):
    # get positions along gradients for text colours
    grad_position = 0.0
    if instances<1: instances = 1
    grad_amount = float(1.00/instances)
    loop = 0
    while loop<=instances:
        if loop == 0:
            list_of_pos = [grad_position]
        else:
            list_of_pos.append(grad_position)
        grad_position += grad_amount
        loop += 1
    return list_of_pos

#######################################################################################################
	
def fn_shade_colour(f_colour, bl_colour, instances, amount):
    red_amount_old = f_colour[0]
    green_amount_old = f_colour[1]
    blue_amount_old = f_colour[2]
    red_amount_new = bl_colour[0]
    green_amount_new = bl_colour[1]
    blue_amount_new = bl_colour[2]
    zero = 0.0
	
    if red_amount_old<red_amount_new:
        red_spread = float(red_amount_new - red_amount_old)
        red_step = float(red_spread/instances)
    else:
        red_spread = float(red_amount_old - red_amount_new)
        red_step = float(zero - (red_spread/instances))
		
    if green_amount_old<green_amount_new:
        green_spread = float(green_amount_new - green_amount_old)
        green_step = float(green_spread/instances)
    else:
        green_spread = float(green_amount_old - green_amount_new)
        green_step = float(zero - float((green_spread/instances)))

    if blue_amount_old<blue_amount_new:
        blue_spread = float(blue_amount_new - blue_amount_old)
        blue_step = float(blue_spread/instances)
    else:
        blue_spread = float(blue_amount_old - blue_amount_new)
        blue_step = float(zero - float((blue_spread/instances)))

    new_red_amount = red_amount_old + (red_step * amount)
    if new_red_amount < 0: new_red_amount = 0
    if new_red_amount > 255: new_red_amount = 255
	
    new_green_amount = green_amount_old + (green_step * amount)
    if new_green_amount < 0: new_green_amount = 0
    if new_green_amount > 255: new_green_amount = 255
	
    new_blue_amount = blue_amount_old + (blue_step * amount)
    if new_blue_amount < 0: new_blue_amount = 0
    if new_blue_amount > 255: new_blue_amount = 255
	
    new_f_colour = (int(new_red_amount), int(new_green_amount), int(new_blue_amount))
	
    return new_f_colour
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################
	
######################################################################################################################################
	
register(
    "python_fu_Path_copy_rotate_shift_plus",
    "Provides options for rotating, moving and stroking user paths",
    "Provides options for rotating, moving and stroking user paths",
    "Brian Hahn, Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2011 - 2020",
    "<Image>/Filters/Path Copy Rotate Shift Plus...",
    "*",
    [
    (PF_SPINNER, "offset_X", "Offset X (px): ", 3, (-1000, 1000, 1)),
    (PF_SPINNER, "offset_Y", "Offset Y (px): ", 3, (-1000, 1000, 1)),
    (PF_SPINNER, "rotation_angle", "Rotation Angle: ", 0.5, (-180.0, 180.0, 0.1)),
    (PF_SPINNER, "rotation_X", "Rotation centre X: ", 500, (0, 5000, 1)),
    (PF_SPINNER, "rotation_Y", "Rotation centre Y: ", 500, (0, 5000, 1)),
    (PF_TOGGLE, "image_centre", "Use image centre: ", 1),
    (PF_SPINNER, "copies", "Number of copies: ", 20, (2, 500, 1)),
    (PF_SPINNER, "path_width", "Path width: ", 2, (1, 500, 1)),
    (PF_SPINNER, "path_modifier", "Path modifier: ", 0.0, (-10.0, 10.0, 0.5)),
    (PF_OPTION, "modifier_options", "Modifier options: ", 0, ("No modifier effect","Apply to x offset", "Apply to y offset", "Apply to rotation", "Apply to x and y offsets", "Apply to x offset and rotation ", "Apply to y offset and rotation", "Apply to all")),
    (PF_OPTION, "colour_options", "Colour option: ", 0, ("Use current foreground colour", "Use current background colour", "Shade colours (FG-BG)", "Shade colours (BG-FG)", "Use current gradient in sequence", "Use current gradient randomly")),
    (PF_OPTION, "path_options", "Path options", 0, ("Keep paths", "Merge paths", "Remove paths")),
    ],
    [],
    do_stuff)

main()
